#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"gpg secrets tests"

import os
import tempfile
import unittest

import kapitan.cached as cached
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from kapitan.refs.base import RefController, RefParams, Revealer
from kapitan.refs.secrets.gpg import GPG_KWARGS, GPG_TARGET_FINGERPRINTS, GPGSecret, gpg_obj

# set GNUPGHOME for test_cli
GNUPGHOME = tempfile.mkdtemp()
os.environ["GNUPGHOME"] = GNUPGHOME

gpg_obj(gnupghome=GNUPGHOME)

KEY = cached.gpg_obj.gen_key(
    cached.gpg_obj.gen_key_input(key_type="RSA", key_length=2048, passphrase="testphrase")
)
KEY2 = cached.gpg_obj.gen_key(
    cached.gpg_obj.gen_key_input(key_type="RSA", key_length=2048, passphrase="testphrase")
)
GPG_TARGET_FINGERPRINTS["KEY"] = KEY.fingerprint
GPG_KWARGS["passphrase"] = "testphrase"

REFS_HOME = tempfile.mkdtemp()
REF_CONTROLLER = RefController(REFS_HOME)
REVEALER = Revealer(REF_CONTROLLER)
REF_CONTROLLER_EMBEDDED = RefController(REFS_HOME, embed_refs=True)
REVEALER_EMBEDDED = Revealer(REF_CONTROLLER_EMBEDDED)


class GPGSecretsTest(unittest.TestCase):
    "Test GPG secrets"

    def test_gpg_write_reveal(self):
        "write secret, confirm secret file exists, reveal and compare content"
        tag = "?{gpg:secret/sauce}"
        REF_CONTROLLER[tag] = GPGSecret("super secret value", [{"fingerprint": KEY.fingerprint}])
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/sauce")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("I am a file with a ?{gpg:secret/sauce}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a super secret value", revealed)

    def test_gpg_write_embedded_reveal(self):
        "write and compile embedded secret, confirm secret file exists, reveal and compare content"
        tag = "?{gpg:secret/sauce}"
        REF_CONTROLLER_EMBEDDED[tag] = GPGSecret("super secret value", [{"fingerprint": KEY.fingerprint}])
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/sauce")))
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("I am a file with a {}".format(ref_obj.compile()))
        revealed = REVEALER_EMBEDDED.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a super secret value", revealed)

    def test_gpg_base64_write_reveal(self):
        """
        write secret for base64 encoded content, confirm secret file exists,
        reveal and compare content
        """
        tag = "?{gpg:secret/sauce2}"
        REF_CONTROLLER[tag] = GPGSecret(
            "super secret value", [{"fingerprint": KEY.fingerprint}], encode_base64=True
        )
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/sauce2")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("I am a file with a ?{gpg:secret/sauce2}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", revealed)

    def test_gpg_base64_write_embedded_reveal(self):
        """
        write and compile embedded secret for base64 encoded content, confirm secret file exists,
        reveal and compare content
        """
        tag = "?{gpg:secret/sauce2}"
        REF_CONTROLLER_EMBEDDED[tag] = GPGSecret(
            "super secret value", [{"fingerprint": KEY.fingerprint}], encode_base64=True
        )
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/sauce2")))
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("I am a file with a {}".format(ref_obj.compile()))
        revealed = REVEALER_EMBEDDED.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", revealed)

    def test_gpg_function_ed25519(self):
        "write ed25519 (private and public), confirm secret file exists, reveal and check"

        tag = "?{gpg:secret/ed25519||ed25519}"
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/ed25519")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("?{gpg:secret/ed25519}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        try:
            serialization.load_pem_private_key(revealed.encode(), password=None, backend=default_backend())
        except ValueError:
            raise Exception("Failed to decode ed25519 private key")

        REVEALER._reveal_tag_without_subvar.cache_clear()
        tag = "?{gpg:secret/ed25519||ed25519}"
        REF_CONTROLLER[tag] = RefParams()
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)

        try:
            private_key = serialization.load_pem_private_key(
                revealed.encode(), password=None, backend=default_backend()
            )
        except ValueError:
            raise Exception("Failed to decode ed25519 private key")

        # Test 'publickey' with previous private key as the parameter
        tag_ed25519public = "?{gpg:secret/ed25519public||reveal:secret/ed25519|publickey}"
        REF_CONTROLLER[tag_ed25519public] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/ed25519")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("?{gpg:secret/ed25519public}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual(revealed.splitlines()[0], "-----BEGIN PUBLIC KEY-----")

    def test_gpg_function_rsa(self):
        "write rsa (private and public), confirm secret file exists, reveal and check"

        tag = "?{gpg:secret/rsa||rsa}"
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/rsa")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("?{gpg:secret/rsa}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        try:
            serialization.load_pem_private_key(revealed.encode(), password=None, backend=default_backend())
        except ValueError:
            raise Exception("Failed to decode RSA private key")

        REVEALER._reveal_tag_without_subvar.cache_clear()
        # Test with parameter key_size=2048
        tag = "?{gpg:secret/rsa||rsa:2048}"
        REF_CONTROLLER[tag] = RefParams()
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)

        try:
            private_key = serialization.load_pem_private_key(
                revealed.encode(), password=None, backend=default_backend()
            )
        except ValueError:
            raise Exception("Failed to decode RSA private key")

        self.assertEqual(private_key.key_size, 2048)

        # Test rsapublic with previous private key as the parameter
        tag_rsapublic = "?{gpg:secret/rsapublic||reveal:secret/rsa|rsapublic}"
        REF_CONTROLLER[tag_rsapublic] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/rsa")))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("?{gpg:secret/rsapublic}")
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual(revealed.splitlines()[0], "-----BEGIN PUBLIC KEY-----")

    def test_gpg_update_recipients(self):
        """
        update existing base64'd secret with another recipient, confirm content is the same
        """
        tag = "?{gpg:secret/sauce_with_fingerprints}"
        secret = GPGSecret("super secret value", [{"fingerprint": KEY.fingerprint}], encode_base64=True)
        REF_CONTROLLER[tag] = secret
        ref = REF_CONTROLLER[tag]
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "secret/sauce_with_fingerprints")))
        self.assertTrue(len(ref.recipients), 1)

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, "w") as fp:
            fp.write("I am a file with a ?{gpg:secret/sauce_with_fingerprints}")

        new_recipients = [{"fingerprint": KEY.fingerprint}, {"fingerprint": KEY2.fingerprint}]
        ref.update_recipients(new_recipients)
        ref = REF_CONTROLLER[tag]

        self.assertTrue(len(ref.recipients), 2)
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", revealed)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import sys
import types

from six import iteritems, next

from collections import namedtuple
from reclass.utils.dictpath import DictPath
from reclass.utils.parameterdict import ParameterDict
from reclass.utils.parameterlist import ParameterList
from reclass.values.value import Value
from reclass.values.valuelist import ValueList
from reclass.errors import InfiniteRecursionError, ResolveError
from reclass.errors import ResolveErrorList, InterpolationError, ParseError
from reclass.errors import BadReferencesError


class Parameters(object):
    '''
    A class to hold nested dictionaries with the following specialities:

      1. "merging" a dictionary (the "new" dictionary) into the current
         Parameters causes a recursive walk of the new dict, during which

         - scalars (incl. tuples) are replaced with the value from the new
           dictionary;
         - lists are extended, not replaced;
         - dictionaries are updated (using dict.update), not replaced;

      2. "interpolating" a dictionary means that values within the dictionary
         can reference other values in the same dictionary. Those references
         are collected during merging and then resolved during interpolation,
         which avoids having to walk the dictionary twice. If a referenced
         value contains references itself, those are resolved first, in
         topological order. Therefore, deep references work. Cyclical
         references cause an error.

    To support these specialities, this class only exposes very limited
    functionality and does not try to be a really mapping object.
    '''

    def __init__(self, mapping, settings, uri, parse_strings=True):
        self._settings = settings
        self._uri = uri
        self._base = ParameterDict(uri=self._uri)
        self._unrendered = None
        self._inv_queries = []
        self.resolve_errors = ResolveErrorList()
        self.needs_all_envs = False
        self._parse_strings = parse_strings
        if mapping is not None:
            # initialise by merging
            self.merge(mapping)

    def __len__(self):
        return len(self._base)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._base)

    def __eq__(self, other):
        return isinstance(other, type(self)) \
                and self._base == other._base \
                and self._settings == other._settings

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def has_inv_query(self):
        return len(self._inv_queries) > 0

    def get_inv_queries(self):
        return self._inv_queries

    def as_dict(self):
        return self._base.copy()

    def _wrap_value(self, value):
        if isinstance(value, (Value, ValueList)):
            return value
        elif isinstance(value, dict):
            return self._wrap_dict(value)
        elif isinstance(value, list):
            return self._wrap_list(value)
        else:
            try:
                return Value(value, self._settings, self._uri,
                             parse_string=self._parse_strings)
            except InterpolationError as e:
                e.context = DictPath(self._settings.delimiter)
                raise

    def _get_wrapped(self, position, value):
        try:
            return self._wrap_value(value)
        except InterpolationError as e:
            e.context.add_ancestor(str(position))
            raise

    def _wrap_list(self, source):
        l = ParameterList(uri=self._uri)
        for (k, v) in enumerate(source):
            l.append(self._get_wrapped(k, v))
        return l

    def _wrap_dict(self, source):
        d = ParameterDict(uri=self._uri)
        for (k, v) in iteritems(source):
            d[k] = self._get_wrapped(k, v)
        return d

    def _update_value(self, cur, new):
        if isinstance(cur, Value):
            values = ValueList(cur, self._settings)
        elif isinstance(cur, ValueList):
            values = cur
        else:
            if isinstance(cur, (ParameterDict, ParameterList)):
                uri = cur.uri
            else:
                uri = self._uri
            values = ValueList(Value(cur, self._settings, uri), self._settings)

        if isinstance(new, Value):
            values.append(new)
        elif isinstance(new, ValueList):
            values.extend(new)
        else:
            if isinstance(new, (ParameterDict, ParameterList)):
                uri = new.uri
            else:
                uri = self._uri
            values.append(Value(new, self._settings, uri,
                                parse_string=self._parse_strings))

        return values

    def _merge_dict(self, cur, new):
        """Merge a dictionary with another dictionary.

        Iterate over keys in new. If this is not an initialization merge and
        the key begins with PARAMETER_DICT_KEY_OVERRIDE_PREFIX, override the
        value of the key in cur. Otherwise deeply merge the contents of the key
        in cur with the contents of the key in _merge_recurse over the item.

        Args:
            cur (dict): Current dictionary
            new (dict): Dictionary to be merged
            initmerge (bool): True if called as part of entity init

        Returns:
            dict: a merged dictionary

        """

        for (key, value) in iteritems(new):
            # check key for "control" preffixes (~,=,...)
            key = str(key)
            if key[0] in self._settings.dict_key_prefixes:
                newkey = key[1:]
                if not isinstance(value, Value):
                    value = Value(value, self._settings, self._uri, parse_string=self._parse_strings)
                if key[0] == self._settings.dict_key_override_prefix:
                    value.overwrite = True
                elif key[0] == self._settings.dict_key_constant_prefix:
                    value.constant = True
                value = self._merge_recurse(cur.get(newkey), value)
                key = newkey
            else:
                value = self._merge_recurse(cur.get(key), value)
            cur[key] = value
        cur.uri = new.uri
        return cur

    def _merge_recurse(self, cur, new):
        """Merge a parameter with another parameter.

        Iterate over keys in new. Call _merge_dict, _update_value
        depending on type.

        Args:
            cur: Current parameter
            new: Parameter to be merged

        Returns:
            merged parameter (Value or ValueList)

        """

        if isinstance(new, dict):
            if cur is None:
                cur = ParameterDict(uri=self._uri)
            if isinstance(cur, dict):
                return self._merge_dict(cur, new)
            else:
                return self._update_value(cur, new)
        else:
            if cur is None:
                return new
            else:
                return self._update_value(cur, new)

    def merge(self, other):
        """Merge function (public edition).

        Call _merge_recurse on self with either another Parameter object or a
        dict (for initialization). Set initmerge if it's a dict.

        Args:
            other (dict or Parameter): Thing to merge with self._base

        Returns:
            None: Nothing

        """

        self._unrendered = None
        if isinstance(other, dict):
            wrapped = self._wrap_dict(other)
        elif isinstance(other, self.__class__):
            wrapped = other._wrap_dict(other._base)
        else:
            raise TypeError('Cannot merge %s objects into %s' % (type(other),
                            self.__class__.__name__))
        self._base = self._merge_recurse(self._base, wrapped)

    def _render_simple_container(self, container, key, value, path):
        if isinstance(value, ValueList):
            if value.is_complex:
                p = path.new_subpath(key)
                self._unrendered[p] = True
                container[key] = value
                if value.has_inv_query:
                    self._inv_queries.append((p, value))
                    if value.needs_all_envs:
                        self.needs_all_envs = True
                return
            else:
                value = value.merge()
        if isinstance(value, Value) and value.is_container():
            value = value.contents
        if isinstance(value, dict):
            container[key] = self._render_simple_dict(value, path.new_subpath(key))
        elif isinstance(value, list):
            container[key] = self._render_simple_list(value, path.new_subpath(key))
        elif isinstance(value, Value):
            if value.is_complex:
                p = path.new_subpath(key)
                self._unrendered[p] = True
                container[key] = value
                if value.has_inv_query:
                    self._inv_queries.append((p, value))
                    if value.needs_all_envs:
                        self.needs_all_envs = True
            else:
                container[key] = value.render(None, None)
        else:
            container[key] = value

    def _render_simple_dict(self, dictionary, path):
        new_dict = {}
        for (key, value) in iteritems(dictionary):
            self._render_simple_container(new_dict, key, value, path)
        return new_dict

    def _render_simple_list(self, item_list, path):
        new_list = [ None ] * len(item_list)
        for n, value in enumerate(item_list):
            self._render_simple_container(new_list, n, value, path)
        return new_list

    def interpolate(self, inventory=None):
        self._initialise_interpolate()
        while len(self._unrendered) > 0:
            # we could use a view here, but this is simple enough:
            # _interpolate_inner removes references from the refs hash after
            # processing them, so we cannot just iterate the dict
            path, v = next(iteritems(self._unrendered))
            self._interpolate_inner(path, inventory)
        if self.resolve_errors.have_errors():
            raise self.resolve_errors

    def initialise_interpolation(self):
        self._unrendered = None
        self._initialise_interpolate()

    def _initialise_interpolate(self):
        if self._unrendered is None:
            self._unrendered = {}
            self._inv_queries = []
            self.needs_all_envs = False
            self.resolve_errors = ResolveErrorList()
            self._base = self._render_simple_dict(self._base,
                 DictPath(self._settings.delimiter))

    def _interpolate_inner(self, path, inventory):
        value = path.get_value(self._base)
        if not isinstance(value, (Value, ValueList)):
            # references to lists and dicts are only deepcopied when merged
            # together so it's possible a value with references in a referenced
            # list or dict has already been visited by _interpolate_inner
            del self._unrendered[path]
            return
        self._unrendered[path] = False
        self._interpolate_references(path, value, inventory)
        new = self._interpolate_render_value(path, value, inventory)
        path.set_value(self._base, new)
        del self._unrendered[path]

    def _interpolate_render_value(self, path, value, inventory):
        try:
            new = value.render(self._base, inventory)
        except ResolveError as e:
            e.context = path
            if self._settings.group_errors:
                self.resolve_errors.add(e)
                new = None
            else:
                raise
        except InterpolationError as e:
            e.context = path
            raise

        if isinstance(new, dict):
            new = self._render_simple_dict(new, path)
        elif isinstance(new, list):
            new = self._render_simple_list(new, path)
        return new

    def _interpolate_references(self, path, value, inventory):
        all_refs = False
        while not all_refs:
            for ref in value.get_references():
                path_from_ref = DictPath(self._settings.delimiter, ref)

                if path_from_ref in self._unrendered:
                    if self._unrendered[path_from_ref] is False:
                        # every call to _interpolate_inner replaces the value of
                        # self._unrendered[path] with False
                        # Therefore, if we encounter False instead of True,
                        # it means that we have already processed it and are now
                        # faced with a cyclical reference.
                        raise InfiniteRecursionError(path, ref, value.uri)
                    else:
                        self._interpolate_inner(path_from_ref, inventory)
                else:
                    # ensure ancestor keys are already dereferenced
                    ancestor = DictPath(self._settings.delimiter)
                    for k in path_from_ref.key_parts():
                        ancestor = ancestor.new_subpath(k)
                        if ancestor in self._unrendered:
                            self._interpolate_inner(ancestor, inventory)
            if value.allRefs:
                all_refs = True
            else:
                # not all references in the value could be calculated previously so
                # try recalculating references with current context and recursively
                # call _interpolate_inner if the number of references has increased
                # Otherwise raise an error
                old = len(value.get_references())
                value.assembleRefs(self._base)
                if old == len(value.get_references()):
                    raise BadReferencesError(value.get_references(), str(path), value.uri)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from reclass.settings import Settings
from reclass.values import item


class CompItem(item.ItemWithReferences):

    type = item.ItemTypes.COMPOSITE

    def merge_over(self, other):
        if (other.type == item.ItemTypes.SCALAR or
                other.type == item.ItemTypes.COMPOSITE):
            return self
        raise RuntimeError('Failed to merge %s over %s' % (self, other))

    def render(self, context, inventory):
        # Preserve type if only one item
        if len(self.contents) == 1:
            return self.contents[0].render(context, inventory)
        # Multiple items
        strings = [str(i.render(context, inventory)) for i in self.contents]
        return "".join(strings)

    def __str__(self):
        return ''.join([str(i) for i in self.contents])



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import distutils.version
import errno
import fcntl
import os
import time

# Squelch warning on centos7 due to upgrading cffi
# see https://github.com/saltstack/salt/pull/39871
import warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    try:
        # NOTE: in some distros pygit2 could require special effort to acquire.
        # It is not a problem per se, but it breaks tests for no real reason.
        # This try block is for keeping tests sane.
        import pygit2
    except ImportError:
        pygit2 = None

from six import iteritems

import reclass.errors
from reclass.storage import ExternalNodeStorageBase
from reclass.storage.yamldata import YamlData

FILE_EXTENSION = ('.yml', '.yaml')
STORAGE_NAME = 'yaml_git'

def path_mangler(inventory_base_uri, nodes_uri, classes_uri):
    if nodes_uri == classes_uri:
        raise errors.DuplicateUriError(nodes_uri, classes_uri)
    return nodes_uri, classes_uri


GitMD = collections.namedtuple('GitMD', ['name', 'path', 'id'], rename=False)


class GitURI(object):

    def __init__(self, dictionary):
        self.repo = None
        self.branch = None
        self.root = None
        self.cache_dir = None
        self.lock_dir = None
        self.pubkey = None
        self.privkey = None
        self.password = None
        self.update(dictionary)

    def update(self, dictionary):
        if 'repo' in dictionary: self.repo = dictionary['repo']
        if 'branch' in dictionary: self.branch = dictionary['branch']
        if 'cache_dir' in dictionary: self.cache_dir = dictionary['cache_dir']
        if 'lock_dir' in dictionary: self.lock_dir = dictionary['lock_dir']
        if 'pubkey' in dictionary: self.pubkey = dictionary['pubkey']
        if 'privkey' in dictionary: self.privkey = dictionary['privkey']
        if 'password' in dictionary: self.password = dictionary['password']
        if 'root' in dictionary:
            if dictionary['root'] is None:
                self.root = None
            else:
                self.root = dictionary['root'].replace('/', '.')

    def __repr__(self):
        return '<{0}: {1} {2} {3}>'.format(self.__class__.__name__, self.repo, self.branch, self.root)


class LockFile():
    def __init__(self, file):
        self._file = file

    def __enter__(self):
        self._fd = open(self._file, 'w+')
        start = time.time()
        while True:
            if (time.time() - start) > 120:
                raise IOError('Timeout waiting to lock file: {0}'.format(self._file))
            try:
                fcntl.flock(self._fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                break
            except IOError as e:
                # raise on unrelated IOErrors
                if e.errno != errno.EAGAIN:
                    raise
                else:
                    time.sleep(0.1)

    def __exit__(self, type, value, traceback):
        self._fd.close()


class GitRepo(object):
    def __init__(self, uri, node_name_mangler, class_name_mangler):
        if pygit2 is None:
            raise errors.MissingModuleError('pygit2')
        self.transport, _, self.url = uri.repo.partition('://')
        self.name = self.url.replace('/', '_')
        self.credentials = None
        self.remotecallbacks = None
        if uri.cache_dir is None:
            self.cache_dir = '{0}/{1}/{2}'.format(os.path.expanduser("~"), '.reclass/cache/git', self.name)
        else:
            self.cache_dir = '{0}/{1}'.format(uri.cache_dir, self.name)
        if uri.lock_dir is None:
            self.lock_file = '{0}/{1}/{2}'.format(os.path.expanduser("~"), '.reclass/cache/lock', self.name)
        else:
            self.lock_file = '{0}/{1}'.format(uri.lock_dir, self.name)
        lock_dir = os.path.dirname(self.lock_file)
        if not os.path.exists(lock_dir):
            os.makedirs(lock_dir)
        self._node_name_mangler = node_name_mangler
        self._class_name_mangler = class_name_mangler
        with LockFile(self.lock_file):
            self._init_repo(uri)
            self._fetch()
        self.branches = self.repo.listall_branches()
        self.files = self.files_in_repo()

    def _init_repo(self, uri):
        if os.path.exists(self.cache_dir):
            self.repo = pygit2.Repository(self.cache_dir)
        else:
            os.makedirs(self.cache_dir)
            self.repo = pygit2.init_repository(self.cache_dir, bare=True)
            self.repo.create_remote('origin', self.url)
        if 'ssh' in self.transport:
            if '@' in self.url:
                user, _, _ = self.url.partition('@')
            else:
                user = 'gitlab'

            if uri.pubkey is not None:
                creds = pygit2.Keypair(user, uri.pubkey, uri.privkey, uri.password)
            else:
                creds = pygit2.KeypairFromAgent(user)

            pygit2_version = pygit2.__version__
            if distutils.version.LooseVersion(pygit2_version) >= distutils.version.LooseVersion('0.23.2'):
                self.remotecallbacks = pygit2.RemoteCallbacks(credentials=creds)
                self.credentials = None
            else:
                self.remotecallbacks = None
                self.credentials = creds

    def _fetch(self):
        origin = self.repo.remotes[0]
        fetch_kwargs = {}
        if self.remotecallbacks is not None:
            fetch_kwargs['callbacks'] = self.remotecallbacks
        if self.credentials is not None:
            origin.credentials = self.credentials
        fetch_results = origin.fetch(**fetch_kwargs)
        remote_branches = self.repo.listall_branches(pygit2.GIT_BRANCH_REMOTE)
        local_branches = self.repo.listall_branches()
        for remote_branch_name in remote_branches:
            _, _, local_branch_name = remote_branch_name.partition('/')
            remote_branch = self.repo.lookup_branch(remote_branch_name, pygit2.GIT_BRANCH_REMOTE)
            if local_branch_name not in local_branches:
                local_branch = self.repo.create_branch(local_branch_name, self.repo[remote_branch.target.hex])
                local_branch.upstream = remote_branch
            else:
                local_branch = self.repo.lookup_branch(local_branch_name)
                if local_branch.target != remote_branch.target:
                    local_branch.set_target(remote_branch.target)

        local_branches = self.repo.listall_branches()
        for local_branch_name in local_branches:
            remote_branch_name = '{0}/{1}'.format(origin.name, local_branch_name)
            if remote_branch_name not in remote_branches:
                local_branch = self.repo.lookup_branch(local_branch_name)
                local.branch.delete()

    def get(self, id):
        return self.repo.get(id)

    def files_in_tree(self, tree, path):
        files = []
        for entry in tree:
            if entry.filemode == pygit2.GIT_FILEMODE_TREE:
                subtree = self.repo.get(entry.id)
                if path == '':
                    subpath = entry.name
                else:
                    subpath = '/'.join([path, entry.name])
                files.extend(self.files_in_tree(subtree, subpath))
            else:
                if path == '':
                   relpath = entry.name
                else:
                   relpath = '/'.join([path, entry.name])
                files.append(GitMD(entry.name, relpath, entry.id))
        return files

    def files_in_branch(self, branch):
        tree = self.repo.revparse_single(branch).tree
        return self.files_in_tree(tree, '')

    def files_in_repo(self):
        ret = {}
        for bname in self.branches:
            branch = {}
            files = self.files_in_branch(bname)
            for file in files:
                if file.name.endswith(FILE_EXTENSION):
                    name = os.path.splitext(file.name)[0]
                    relpath = os.path.dirname(file.path)
                    if callable(self._class_name_mangler):
                        relpath, name = self._class_name_mangler(relpath, name)
                    if name in ret:
                        raise reclass.errors.DuplicateNodeNameError(self.url + ' - ' + bname, name, ret[name], file)
                    else:
                        branch[name] = file
            ret[bname] = branch
        return ret

    def nodes(self, branch, subdir):
        ret = {}
        for (name, file) in iteritems(self.files[branch]):
            if subdir is None or name.startswith(subdir):
                node_name = os.path.splitext(file.name)[0]
                relpath = os.path.dirname(file.path)
                if callable(self._node_name_mangler):
                    relpath, node_name = self._node_name_mangler(relpath, node_name)
                if node_name in ret:
                    raise reclass.errors.DuplicateNodeNameError(self.url, name, ret[node_name].path, file.path)
                else:
                    ret[node_name] = file
        return ret


class ExternalNodeStorage(ExternalNodeStorageBase):
    def __init__(self, nodes_uri, classes_uri, compose_node_name):
        super(ExternalNodeStorage, self).__init__(STORAGE_NAME, compose_node_name)
        self._repos = dict()

        if nodes_uri is not None:
            self._nodes_uri = GitURI({ 'branch': 'master' })
            self._nodes_uri.update(nodes_uri)
            self._load_repo(self._nodes_uri)
            self._nodes = self._repos[self._nodes_uri.repo].nodes(self._nodes_uri.branch, self._nodes_uri.root)

        if classes_uri is not None:
            self._classes_default_uri = GitURI({ 'branch': '__env__' })
            self._classes_default_uri.update(classes_uri)
            self._load_repo(self._classes_default_uri)

            self._classes_uri = []
            if 'env_overrides' in classes_uri:
                for override in classes_uri['env_overrides']:
                    for (env, options) in iteritems(override):
                        uri = GitURI(self._classes_default_uri)
                        uri.update({ 'branch': env })
                        uri.update(options)
                        self._classes_uri.append((env, uri))
                        self._load_repo(uri)

            self._classes_uri.append(('*', self._classes_default_uri))

    nodes_uri = property(lambda self: self._nodes_uri)
    classes_uri = property(lambda self: self._classes_uri)

    def get_node(self, name, settings):
        file = self._nodes[name]
        blob = self._repos[self._nodes_uri.repo].get(file.id)
        uri = 'git_fs://{0} {1} {2}'.format(self._nodes_uri.repo, self._nodes_uri.branch, file.path)
        pathname = os.path.splitext(file.path)[0]
        entity = YamlData.from_string(blob.data, uri).get_entity(name, pathname, settings)
        return entity

    def get_class(self, name, environment, settings):
        uri = self._env_to_uri(environment)
        if uri.root is not None:
            name = '{0}.{1}'.format(uri.root, name)
        if uri.repo not in self._repos:
            raise reclass.errors.NotFoundError("Repo " + uri.repo + " unknown or missing")
        if uri.branch not in self._repos[uri.repo].files:
            raise reclass.errors.NotFoundError("Branch " + uri.branch + " missing from " + uri.repo)
        if name not in self._repos[uri.repo].files[uri.branch]:
            raise reclass.errors.NotFoundError("File " + name + " missing from " + uri.repo + " branch " + uri.branch)
        file = self._repos[uri.repo].files[uri.branch][name]
        blob = self._repos[uri.repo].get(file.id)
        uri = 'git_fs://{0} {1} {2}'.format(uri.repo, uri.branch, file.path)
        pathname = os.path.splitext(file.path)[0]
        entity = YamlData.from_string(blob.data, uri).get_entity(name, pathname, settings)
        return entity

    def enumerate_nodes(self):
        return self._nodes.keys()

    def _load_repo(self, uri):
        if uri.repo not in self._repos:
            self._repos[uri.repo] = GitRepo(uri, self.node_name_mangler, self.class_name_mangler)

    def _env_to_uri(self, environment):
        ret = None
        if environment is None:
            ret = self._classes_default_uri
        else:
            for env, uri in self._classes_uri:
                if env == environment:
                    ret = uri
                    break
        if ret is None:
            ret = self._classes_default_uri
        if ret.branch == '__env__':
            ret.branch = environment
        if ret.branch == None:
            ret.branch = 'master'
        return ret



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.settings import Settings
from reclass.values import item


class ScaItem(item.Item):

    type = item.ItemTypes.SCALAR

    def __init__(self, value, settings):
        super(ScaItem, self).__init__(value, settings)

    def merge_over(self, other):
        if other.type in [item.ItemTypes.SCALAR, item.ItemTypes.COMPOSITE]:
            return self
        raise RuntimeError('Failed to merge %s over %s' % (self, other))

    def render(self, context, inventory):
        return self.contents

    def __str__(self):
        return str(self.contents)



import logging
import os
from kapitan.errors import HelmBindingUnavailableError, HelmTemplateError
from kapitan.inputs.base import InputType, CompiledFile
import tempfile
import platform
import yaml

try:
    from kapitan.inputs.helm.helm_binding import ffi
except ImportError:
    pass  # make this feature optional

logger = logging.getLogger(__name__)


class Helm(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("helm", compile_path, search_paths, ref_controller)
        self.helm_values_file = None
        self.helm_values_files = []
        self.helm_params = {}
        self.lib = self.initialise_binding()

    def initialise_binding(self):
        """returns the dl_opened library (.so file) if exists, otherwise None"""
        if platform.system() not in ("Linux", "Darwin"):  # TODO: later add binding for Mac
            return None
        # binding_path is kapitan/inputs/helm/libtemplate.so
        binding_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "libtemplate.so")
        if not os.path.exists(binding_path):
            logger.debug("The helm binding does not exist at {}".format(binding_path))
            return None
        try:
            lib = ffi.dlopen(binding_path)
        except (NameError, OSError) as e:
            raise HelmBindingUnavailableError(
                "There was an error opening helm binding. " "Refer to the exception below:\n" + str(e)
            )
        return lib

    def dump_helm_values(self, helm_values):
        """dump helm values into a yaml file whose path will be passed over to Go helm code"""
        _, self.helm_values_file = tempfile.mkstemp(".helm_values.yml", text=True)
        with open(self.helm_values_file, "w") as fp:
            yaml.safe_dump(helm_values, fp)

    def set_helm_values_files(self, helm_values_files):
        self.helm_values_files = helm_values_files

    def set_helm_params(self, helm_params):
        self.helm_params = helm_params

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Render templates in file_path/templates and write to compile_path.
        file_path must be a directory containing helm chart.
        kwargs:
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
        """
        if not self.lib:
            raise HelmBindingUnavailableError(
                "Helm binding is not supported for {}."
                "\nOr the binding does not exist.".format(platform.system())
            )

        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)

        temp_dir = tempfile.mkdtemp()
        os.makedirs(os.path.dirname(compile_path), exist_ok=True)
        # save the template output to temp dir first
        error_message = self.render_chart(
            chart_dir=file_path,
            output_path=temp_dir,
            helm_values_file=self.helm_values_file,
            helm_values_files=self.helm_values_files,
            **self.helm_params,
        )
        if error_message:
            raise HelmTemplateError(error_message)

        walk_root_files = os.walk(temp_dir)
        for current_dir, _, files in walk_root_files:
            for file in files:  # go through all the template files
                rel_dir = os.path.relpath(current_dir, temp_dir)
                rel_file_name = os.path.join(rel_dir, file)
                full_file_name = os.path.join(current_dir, file)
                with open(full_file_name, "r") as f:
                    item_path = os.path.join(compile_path, rel_file_name)
                    os.makedirs(os.path.dirname(item_path), exist_ok=True)
                    with CompiledFile(
                        item_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name
                    ) as fp:
                        yml_obj = list(yaml.safe_load_all(f))
                        fp.write_yaml(yml_obj)
                        logger.debug("Wrote file %s to %s", full_file_name, item_path)

        self.helm_values_file = None  # reset this
        self.helm_params = {}

    def default_output_type(self):
        return None

    def render_chart(self, chart_dir, output_path, **kwargs):
        """renders helm chart located at chart_dir, and stores the output to output_path"""
        if kwargs.get("helm_values_file", None):
            helm_values_file = ffi.new("char[]", kwargs["helm_values_file"].encode("ascii"))
        else:
            # the value in kwargs can be None
            helm_values_file = ffi.new("char[]", "".encode("ascii"))

        if kwargs.get("namespace", None):
            namespace = ffi.new("char[]", kwargs["namespace"].encode("ascii"))
        else:
            namespace = ffi.new("char[]", "default".encode("ascii"))

        if kwargs.get("release_name", None):
            release_name = ffi.new("char[]", kwargs["release_name"].encode("ascii"))
        else:
            release_name = ffi.new("char[]", "".encode("ascii"))

        if kwargs.get("name_template", None):
            name_template = ffi.new("char[]", kwargs["name_template"].encode("ascii"))
        else:
            name_template = ffi.new("char[]", "".encode("ascii"))

        char_dir_buf = ffi.new("char[]", chart_dir.encode("ascii"))
        output_path_buf = ffi.new("char[]", output_path.encode("ascii"))

        helm_values_files = []
        if kwargs.get("helm_values_files", []):
            for file_name in kwargs["helm_values_files"]:
                helm_values_files.append(ffi.new("char[]", file_name.encode("ascii")))

        helm_values_files_len = ffi.cast("int", len(helm_values_files))
        helm_values_files = ffi.new("char *[]", helm_values_files)

        c_error_message = self.lib.renderChart(
            char_dir_buf,
            output_path_buf,
            helm_values_file,
            namespace,
            release_name,
            name_template,
            helm_values_files,
            helm_values_files_len,
        )
        error_message = ffi.string(c_error_message)  # this creates a copy as bytes
        self.lib.free(c_error_message)  # free the char* returned by go
        return error_message.decode("utf-8")  # empty if no error



from cffi import FFI


def main():
    ffi = FFI()

    # declare functions to export
    ffi.cdef(
        """
        char* renderChart(char* p0, char* p1, char* p2, char* p3, char* p4, char* p5, char** p6, int p7);
        void free(void* ptr);
    """
    )

    ffi.set_source("helm_binding", None)  # specify name for importing this module

    ffi.compile(verbose=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"copy tests"
import filecmp
import hashlib
import os
import shutil
import sys
import unittest

from kapitan.cli import main
from kapitan.inputs.copy import Copy
from kapitan.utils import directory_hash

search_path = ""
ref_controller = ""
test_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_copy")
compile_path = os.path.join(test_path, "output")
file_path = os.path.join(test_path, "input")
test_file_path = os.path.join(file_path, "test_copy_input")
test_file_compiled_path = os.path.join(compile_path, "test_copy_input")
test_file_content = """
apiVersion: v1
kind: Pod
metadata:
  name: alpine
  namespace: default
spec:
  containers:
  - image: alpine:3.2
    command:
      - /bin/sh
      - "-c"
      - "sleep 60m"
    imagePullPolicy: IfNotPresent
    name: alpine
  restartPolicy: Always
"""


def test_dirs_bootstrap_helper():
    for folder in [file_path, compile_path]:
        os.makedirs(folder, exist_ok=True)
        with open(test_file_path, "w") as f:
            f.write(test_file_content)


class CopyTest(unittest.TestCase):
    def setUp(self):
        try:
            shutil.rmtree(test_path)
        except FileNotFoundError:
            pass

        self.copy_compiler = Copy(compile_path, search_path, ref_controller)

    def test_copy_file_folder(self):
        test_dirs_bootstrap_helper()
        self.copy_compiler.compile_file(test_file_path, compile_path, None)
        self.test_file_hash = hashlib.sha1(test_file_content.encode()).digest()
        with open(test_file_compiled_path) as f:
            test_file_compiled_hash = hashlib.sha1(f.read().encode()).digest()
            self.assertEqual(self.test_file_hash, test_file_compiled_hash)

    def test_copy_folder_folder(self):
        test_dirs_bootstrap_helper()
        self.copy_compiler.compile_file(file_path, compile_path, None)
        file_path_hash = directory_hash(file_path)
        compile_path_hash = directory_hash(compile_path)
        self.assertEqual(file_path_hash, compile_path_hash)

    def tearDown(self):
        try:
            shutil.rmtree(test_path)
        except FileNotFoundError:
            pass


class CompileCopyTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join(os.getcwd(), "examples", "kubernetes"))

    def validate_files_were_copied(self):
        original_filepath = os.path.join("components", "busybox", "pod.yml")
        copied_filepath = os.path.join("compiled", "busybox", "copy", "pod.yml")
        self.assertTrue(filecmp.cmp(original_filepath, copied_filepath))

        original_filepath = os.path.join("copy_target")
        copied_filepath = os.path.join("compiled", "busybox", "copy", "copy_target")
        self.assertTrue(filecmp.cmp(original_filepath, copied_filepath))

        original_filepath = os.path.join("copy_target")
        copied_filepath = os.path.join("compiled", "busybox", "copy_target")
        self.assertTrue(filecmp.cmp(original_filepath, copied_filepath))

    def test_compiled_copy_target(self):
        sys.argv = ["kapitan", "compile", "-t", "busybox"]
        main()
        self.validate_files_were_copied()

    def test_compiled_copy_all_targets(self):
        sys.argv = ["kapitan", "compile"]
        main()
        self.validate_files_were_copied()

    def tearDown(self):
        os.chdir(os.path.join(os.getcwd(), "..", ".."))



from reclass.settings import Settings
from reclass.values.value import Value
from reclass.values.compitem import CompItem
from reclass.values.scaitem import ScaItem
from reclass.values.valuelist import ValueList
from reclass.values.listitem import ListItem
from reclass.values.dictitem import DictItem
import unittest

SETTINGS = Settings()

class TestCompItem(unittest.TestCase):

    def test_assembleRefs_no_items(self):
        composite = CompItem([], SETTINGS)

        self.assertFalse(composite.has_references)

    def test_assembleRefs_one_item_without_refs(self):
        val1 = Value('foo',  SETTINGS, '')

        composite = CompItem([val1], SETTINGS)

        self.assertFalse(composite.has_references)

    def test_assembleRefs_one_item_with_one_ref(self):
        val1 = Value('${foo}',  SETTINGS, '')
        expected_refs = ['foo']

        composite = CompItem([val1], SETTINGS)

        self.assertTrue(composite.has_references)
        self.assertEquals(composite.get_references(), expected_refs)

    def test_assembleRefs_one_item_with_two_refs(self):
        val1 = Value('${foo}${bar}',  SETTINGS, '')
        expected_refs = ['foo', 'bar']

        composite = CompItem([val1], SETTINGS)

        self.assertTrue(composite.has_references)
        self.assertEquals(composite.get_references(), expected_refs)

    def test_assembleRefs_two_items_one_with_one_ref_one_without(self):
        val1 = Value('${foo}bar',  SETTINGS, '')
        val2 = Value('baz',  SETTINGS, '')
        expected_refs = ['foo']

        composite = CompItem([val1, val2], SETTINGS)

        self.assertTrue(composite.has_references)
        self.assertEquals(composite.get_references(), expected_refs)

    def test_assembleRefs_two_items_both_with_one_ref(self):
        val1 = Value('${foo}',  SETTINGS, '')
        val2 = Value('${bar}',  SETTINGS, '')
        expected_refs = ['foo', 'bar']

        composite = CompItem([val1, val2], SETTINGS)

        self.assertTrue(composite.has_references)
        self.assertEquals(composite.get_references(), expected_refs)

    def test_assembleRefs_two_items_with_two_refs(self):
        val1 = Value('${foo}${baz}',  SETTINGS, '')
        val2 = Value('${bar}${meep}',  SETTINGS, '')
        expected_refs = ['foo', 'baz', 'bar', 'meep']

        composite = CompItem([val1, val2], SETTINGS)

        self.assertTrue(composite.has_references)
        self.assertEquals(composite.get_references(), expected_refs)

    def test_string_representation(self):
        composite = CompItem(Value(1, SETTINGS, ''), SETTINGS)
        expected = '1'

        result = str(composite)

        self.assertEquals(result, expected)

    def test_render_single_item(self):
        val1 = Value('${foo}',  SETTINGS, '')

        composite = CompItem([val1], SETTINGS)

        self.assertEquals(1, composite.render({'foo': 1}, None))


    def test_render_multiple_items(self):
        val1 = Value('${foo}',  SETTINGS, '')
        val2 = Value('${bar}',  SETTINGS, '')

        composite = CompItem([val1, val2], SETTINGS)

        self.assertEquals('12', composite.render({'foo': 1, 'bar': 2}, None))

    def test_merge_over_merge_scalar(self):
        val1 = Value(None, SETTINGS, '')
        scalar = ScaItem(1, SETTINGS)
        composite = CompItem([val1], SETTINGS)

        result = composite.merge_over(scalar)

        self.assertEquals(result, composite)

    def test_merge_over_merge_composite(self):
        val1 = Value(None, SETTINGS, '')
        val2 = Value(None, SETTINGS, '')
        composite1 = CompItem([val1], SETTINGS)
        composite2 = CompItem([val2], SETTINGS)

        result = composite2.merge_over(composite1)

        self.assertEquals(result, composite2)

    def test_merge_other_types_not_allowed(self):
        other = type('Other', (object,), {'type': 34})
        val1 = Value(None, SETTINGS, '')
        composite = CompItem([val1], SETTINGS)

        self.assertRaises(RuntimeError, composite.merge_over, other)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import base64
import logging
import os
import sys

from kapitan.errors import KapitanError, RefHashMismatchError
from kapitan.refs.base import PlainRef, RefController, Revealer
from kapitan.refs.base64 import Base64Ref
from kapitan.refs.env import EnvRef
from kapitan.refs.secrets.awskms import AWSKMSSecret
from kapitan.refs.secrets.gkms import GoogleKMSSecret
from kapitan.refs.secrets.gpg import GPGSecret, lookup_fingerprints
from kapitan.refs.secrets.vaultkv import VaultSecret
from kapitan.resources import inventory_reclass
from kapitan.utils import fatal_error, search_target_token_paths

logger = logging.getLogger(__name__)


def handle_refs_command(args):
    ref_controller = RefController(args.refs_path)

    if args.write is not None:
        ref_write(args, ref_controller)
    elif args.reveal:
        ref_reveal(args, ref_controller)
    elif args.update:
        secret_update(args, ref_controller)
    elif args.update_targets or args.validate_targets:
        secret_update_validate(args, ref_controller)


def ref_write(args, ref_controller):
    "Write ref to ref_controller based on cli args"
    token_name = args.write
    file_name = args.file
    data = None

    if file_name is None:
        fatal_error("--file is required with --write")
    if file_name == "-":
        data = ""
        for line in sys.stdin:
            data += line
    else:
        with open(file_name) as fp:
            data = fp.read()

    if token_name.startswith("gpg:"):
        type_name, token_path = token_name.split(":")
        recipients = [dict((("name", name),)) for name in args.recipients]
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError(
                    "parameters.kapitan.secrets not defined in inventory of target {}".format(
                        args.target_name
                    )
                )

            recipients = kap_inv_params["secrets"]["gpg"]["recipients"]
        if not recipients:
            raise KapitanError(
                "No GPG recipients specified. Use --recipients or specify them in "
                + "parameters.kapitan.secrets.gpg.recipients and use --target-name"
            )
        secret_obj = GPGSecret(data, recipients, encode_base64=args.base64)
        tag = "?{{gpg:{}}}".format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("gkms:"):
        type_name, token_path = token_name.split(":")
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError(
                    "parameters.kapitan.secrets not defined in inventory of target {}".format(
                        args.target_name
                    )
                )

            key = kap_inv_params["secrets"]["gkms"]["key"]
        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target-name"
            )
        secret_obj = GoogleKMSSecret(data, key, encode_base64=args.base64)
        tag = "?{{gkms:{}}}".format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("awskms:"):
        type_name, token_path = token_name.split(":")
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError(
                    "parameters.kapitan.secrets not defined in inventory of target {}".format(
                        args.target_name
                    )
                )

            key = kap_inv_params["secrets"]["awskms"]["key"]
        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target-name"
            )
        secret_obj = AWSKMSSecret(data, key, encode_base64=args.base64)
        tag = "?{{awskms:{}}}".format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("base64:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = Base64Ref(_data, encoding=encoding)
        tag = "?{{base64:{}}}".format(token_path)
        ref_controller[tag] = ref_obj

    elif token_name.startswith("vaultkv:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        vault_params = {}
        encoding = "original"
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError(
                    "parameters.kapitan.secrets not defined in inventory of target {}".format(
                        args.target_name
                    )
                )

            vault_params = kap_inv_params["secrets"]["vaultkv"]
        if args.vault_auth:
            vault_params["auth"] = args.vault_auth
        if vault_params.get("auth") is None:
            raise KapitanError(
                "No Authentication type parameter specified. Specify it"
                " in parameters.kapitan.secrets.vaultkv.auth and use --target-name or use --vault-auth"
            )

        secret_obj = VaultSecret(_data, vault_params)
        tag = "?{{vaultkv:{}}}".format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("plain:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = PlainRef(_data, encoding=encoding)
        tag = "?{{plain:{}}}".format(token_path)
        ref_controller[tag] = ref_obj

    elif token_name.startswith("env:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = EnvRef(_data, encoding=encoding)
        tag = "?{{env:{}}}".format(token_path)
        ref_controller[tag] = ref_obj

    else:
        fatal_error(
            "Invalid token: {name}. Try using gpg/gkms/awskms/vaultkv/base64/plain/env:{name}".format(
                name=token_name
            )
        )


def secret_update(args, ref_controller):
    "Update secret gpg recipients/gkms/awskms key"
    # TODO --update *might* mean something else for other types
    token_name = args.update
    if token_name.startswith("gpg:"):
        # args.recipients is a list, convert to recipients dict
        recipients = [dict([("name", name),]) for name in args.recipients]
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            recipients = kap_inv_params["secrets"]["gpg"]["recipients"]
        if not recipients:
            raise KapitanError(
                "No GPG recipients specified. Use --recipients or specify them in "
                + "parameters.kapitan.secrets.gpg.recipients and use --target"
            )
        type_name, token_path = token_name.split(":")
        tag = "?{{gpg:{}}}".format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_recipients(recipients)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("gkms:"):
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params["secrets"]["gkms"]["key"]
        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target"
            )
        type_name, token_path = token_name.split(":")
        tag = "?{{gkms:{}}}".format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("awskms:"):
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv["nodes"][args.target_name]["parameters"]["kapitan"]
            if "secrets" not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params["secrets"]["awskms"]["key"]
        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target"
            )
        type_name, token_path = token_name.split(":")
        tag = "?{{awskms:{}}}".format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    else:
        fatal_error("Invalid token: {name}. Try using gpg/gkms/awskms:{name}".format(name=token_name))


def ref_reveal(args, ref_controller):
    "Reveal secrets in file_name"
    revealer = Revealer(ref_controller)
    file_name = args.file
    reffile_name = args.ref_file
    tag_name = args.tag

    if file_name is None and reffile_name is None and tag_name is None:
        fatal_error("--file or --ref-file is required with --reveal")
    try:
        if file_name == "-" or reffile_name == "-":
            out = revealer.reveal_raw_file(None)
            sys.stdout.write(out)
        elif file_name:
            for rev_obj in revealer.reveal_path(file_name):
                sys.stdout.write(rev_obj.content)
        elif reffile_name:
            ref = ref_controller.ref_from_ref_file(reffile_name)
            sys.stdout.write(ref.reveal())
        elif tag_name:
            out = revealer.reveal_raw_string(tag_name)
            sys.stdout.write(out)
    except (RefHashMismatchError, KeyError):
        raise KapitanError("Reveal failed for file {name}".format(name=file_name))


def secret_update_validate(args, ref_controller):
    "Validate and/or update target secrets"
    # update gpg recipients/gkms/awskms key for all secrets in secrets_path
    # use --refs-path to set scanning path
    inv = inventory_reclass(args.inventory_path)
    targets = set(inv["nodes"].keys())
    secrets_path = os.path.abspath(args.refs_path)
    target_token_paths = search_target_token_paths(secrets_path, targets)
    ret_code = 0

    for target_name, token_paths in target_token_paths.items():
        kap_inv_params = inv["nodes"][target_name]["parameters"]["kapitan"]
        if "secrets" not in kap_inv_params:
            raise KapitanError("parameters.kapitan.secrets not defined in {}".format(target_name))

        try:
            recipients = kap_inv_params["secrets"]["gpg"]["recipients"]
        except KeyError:
            recipients = None
        try:
            gkey = kap_inv_params["secrets"]["gkms"]["key"]
        except KeyError:
            gkey = None
        try:
            awskey = kap_inv_params["secrets"]["awskms"]["key"]
        except KeyError:
            awskey = None
        try:
            vaultkv = kap_inv_params["secrets"]["vaultkv"]["auth"]
        except KeyError:
            vaultkv = None

        for token_path in token_paths:
            if token_path.startswith("?{gpg:"):
                if not recipients:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory gpg recipients, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                secret_obj = ref_controller[token_path]
                target_fingerprints = set(lookup_fingerprints(recipients))
                secret_fingerprints = set(lookup_fingerprints(secret_obj.recipients))
                if target_fingerprints != secret_fingerprints:
                    if args.validate_targets:
                        logger.info("%s recipient mismatch", token_path)
                        to_remove = secret_fingerprints.difference(target_fingerprints)
                        to_add = target_fingerprints.difference(secret_fingerprints)
                        if to_remove:
                            logger.info("%s needs removal", to_remove)
                        if to_add:
                            logger.info("%s needs addition", to_add)
                        ret_code = 1
                    else:
                        new_recipients = [dict([("fingerprint", f),]) for f in target_fingerprints]
                        secret_obj.update_recipients(new_recipients)
                        ref_controller[token_path] = secret_obj

            elif token_path.startswith("?{gkms:"):
                if not gkey:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory gkms key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                secret_obj = ref_controller[token_path]
                if gkey != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(gkey)
                        ref_controller[token_path] = secret_obj

            elif token_path.startswith("?{awskms:"):
                if not awskey:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory awskms key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                secret_obj = ref_controller[token_path]
                if awskey != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(awskey)
                        ref_controller[token_path] = secret_obj

            else:
                logger.info("Invalid secret %s, could not get type, skipping", token_path)

    sys.exit(ret_code)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from importlib import import_module

class StorageBackendLoader(object):

    def __init__(self, storage_name):
        self._name = str('reclass.storage.' + storage_name)
        try:
            self._module = import_module(self._name)
        except ImportError as e:
            raise NotImplementedError

    def load(self, klassname='ExternalNodeStorage'):
        klass = getattr(self._module, klassname, None)
        if klass is None:
            raise AttributeError('Storage backend class {0} does not export '
                                 '"{1}"'.format(self._name, klassname))
        return klass

    def path_mangler(self, name='path_mangler'):
        function = getattr(self._module, name, None)
        if function is None:
            raise AttributeError('Storage backend class {0} does not export '
                                 '"{1}"'.format(self._name, name))
        return function



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"inventory tests"

import unittest

from kapitan.resources import inventory


class InventoryTargetTest(unittest.TestCase):
    def test_inventory_target(self):
        inv = inventory(["examples/kubernetes"], "minikube-es")
        self.assertEqual(inv["parameters"]["cluster"]["name"], "minikube")

    def test_inventory_all_targets(self):
        inv = inventory(["examples/kubernetes"], None)
        self.assertNotEqual(inv.get("minikube-es"), None)



from reclass.settings import Settings
from reclass.values.value import Value
from reclass.values.compitem import CompItem
from reclass.values.scaitem import ScaItem
from reclass.values.valuelist import ValueList
from reclass.values.listitem import ListItem
from reclass.values.dictitem import DictItem
import unittest

SETTINGS = Settings()

class TestScaItem(unittest.TestCase):

    def test_merge_over_merge_scalar(self):
        scalar1 = ScaItem([1], SETTINGS)
        scalar2 = ScaItem([2], SETTINGS)

        result = scalar2.merge_over(scalar1)

        self.assertEquals(result.contents, scalar2.contents)

    def test_merge_over_merge_composite(self):
        scalar1 = CompItem(Value(1, SETTINGS, ''), SETTINGS)
        scalar2 = ScaItem([2], SETTINGS)

        result = scalar2.merge_over(scalar1)

        self.assertEquals(result.contents, scalar2.contents)

    def test_merge_other_types_not_allowed(self):
        other = type('Other', (object,), {'type': 34})
        val1 = Value(None, SETTINGS, '')
        scalar = ScaItem(val1, SETTINGS)

        self.assertRaises(RuntimeError, scalar.merge_over, other)

if __name__ == '__main__':
    unittest.main()
